from dataclasses import dataclass
import os
import pathlib
import tomllib
import shutil
from typing import Any, Optional

from cursedtodo.config.arguments import Arguments
from cursedtodo.models.calendar import Calendar


@dataclass
class UIConfig:
    window_name: str
    show_footer_keybindings: bool
    select_first: bool
    rounded_borders: bool
    date_format: str
    default_calendar: Optional[str] = None
    category_colors: bool = False
    confirm_mark_as_done: bool = False


@dataclass
class Columns:
    property: str
    width: int


@dataclass
class _Config:
    calendars: list[Calendar]
    ui: UIConfig
    columns: list[Columns]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "_Config":
        calendars: list[Calendar] = [
            Calendar(i, **cal) for i, cal in enumerate(data.get("calendars", {}))
        ]
        ui = UIConfig(**data.get("ui", {}))
        columns: list[Columns] = [Columns(**col) for col in data.get("columns", {})]
        default_calendar = next(filter(lambda cal: cal.default, calendars))
        ui.default_calendar = (
            default_calendar.name if default_calendar is not None else None
        )
        # raise Exception(calendars, data)
        return cls(calendars=calendars, ui=ui, columns=columns)


def _init_config() -> _Config:
    print("Parsing configuration file...")
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME") or os.path.join(
        os.path.expanduser("~"), ".config"
    )
    config_file_path = Arguments.config or os.path.join(
        xdg_config_home, "cursedtodo/config.toml"
    )
    config_file = pathlib.Path(config_file_path)

    if not config_file.exists():
        if Arguments.config is None:
            print(
                "Configuration file not found, creating default configuration file..."
            )
            default_config_file = pathlib.Path(
                os.path.join(
                    pathlib.Path(__file__).parent.resolve(), "default_config.toml"
                )
            )
            shutil.copy(default_config_file, config_file)
        else:
            raise Exception(f"Configuration file not found: {config_file}")

    with config_file.open("rb") as file:
        data = tomllib.load(file)

    return _Config.from_dict(data)



from curses import A_NORMAL, A_STANDOUT, curs_set, window
from typing import Callable


class Button:
    def __init__(
        self,
        window: window,
        y: int,
        x: int,
        name: str,
        action: Callable[[], bool],
        validator: Callable[[int | str], int | str],
    ) -> None:
        self.window = window
        self.x = x
        self.y = y
        self.name = name
        self.action = action
        self.validator = validator

    def render(self, y: int, x: int) -> None:
        self.y = y
        self.x = x
        self.window.addstr(y, x, self.name)

    def focus(self) -> bool | None:
        curs_set(0)
        while True:
            self.window.chgat(self.y, self.x, len(self.name), A_STANDOUT)
            k = self.window.getch()
            self.validator(k)
            if k == 10:
                return self.action()
            if k == 9:
                self.window.chgat(self.y, self.x, A_NORMAL)
                self.window.refresh()
                break
        return None



from collections.abc import Callable
from curses import A_BOLD, KEY_RESIZE, window

from cursedtodo.utils.colors import WHITE
from cursedtodo.utils.editor import Editor
from cursedtodo.views.form.base_field import BaseField


class TextArea(BaseField):
    def __init__(
        self,
        y: int,
        window: window,
        name: str,
        id: str,
        validator: Callable[[int | str], int | str],
        value: str | None = None,
    ):
        super().__init__(y, window, name, id, validator, value)
        # TODO: Make that fixed height dynamic
        self.textwindow = window.derwin(7, 100, y + 1, 1)
        self.textwindow.bkgd(" ", WHITE)
        self.value: str = str(value)
        self.validator = validator
        self.editor = Editor(self.textwindow, self.value or "", self._validator)

    def _validator(self, ch: str | int) -> str | int:
        if ch == KEY_RESIZE:
            self.value = self.editor.gather()
        self.validator(ch)
        return ch

    def render(self) -> None:
        self.window.addstr(self.y, 1, f"{self.name}: ", A_BOLD)
        self.textwindow.move(0, 0)
        self.editor.set_value(self.value or "")
        self.editor.render()
        # Editor().main(self.textwindow,self.value or "")

    def focus(self) -> None:
        self.editor.main()
        self.value = self.editor.gather()



import curses
from typing import Callable, Tuple, Union


class Buffer:
    def __init__(self, lines: str | None):
        self.lines = lines.split() if lines is not None else [""]

    def set_value(self, lines: str) -> None:
        self.lines = lines.splitlines()

    def __len__(self) -> int:
        return len(self.lines)

    def __getitem__(self, index: Union[int, slice]) -> str | list[str]:
        return self.lines[index]

    @property
    def bottom(self) -> int:
        return len(self) - 1

    def insert(self, cursor: "Cursor", string: str) -> None:
        row, col = cursor.row, cursor.col
        if len(self.lines) > 0:
            current = self.lines.pop(row)
            new = current[:col] + string + current[col:]
        else:
            new = string
        self.lines.insert(row, new)

    def split(self, cursor: "Cursor") -> None:
        row, col = cursor.row, cursor.col
        current = self.lines.pop(row)
        self.lines.insert(row, current[:col])
        self.lines.insert(row + 1, current[col:])

    def delete(self, cursor: "Cursor") -> None:
        row, col = cursor.row, cursor.col
        if (row, col) < (self.bottom, len(self[row])):
            current = self.lines.pop(row)
            if col < len(current):
                new = current[:col] + current[col + 1 :]
                self.lines.insert(row, new)
            else:
                next = self.lines.pop(row)
                new = current + next
                self.lines.insert(row, new)


class Cursor:
    def __init__(self, row: int = 0, col: int = 0, col_hint: int | None = None) -> None:
        self.row = row
        self._col = col
        self._col_hint = col if col_hint is None else col_hint

    @property
    def col(self) -> int:
        return self._col

    @col.setter
    def col(self, col: int) -> None:
        self._col = col
        self._col_hint = col

    def up(self, buffer: Buffer) -> None:
        if self.row > 0:
            self.row -= 1
            self._clamp_col(buffer)

    def down(self, buffer: Buffer) -> None:
        if self.row < buffer.bottom:
            self.row += 1
            self._clamp_col(buffer)

    def left(self, buffer: Buffer) -> None:
        if self.col > 0:
            self.col -= 1
        elif self.row > 0:
            self.row -= 1
            self.col = len(buffer[self.row])

    def right(self, buffer: Buffer) -> None:
        if self.col < len(buffer[self.row]):
            self.col += 1
        elif self.row < buffer.bottom:
            self.row += 1
            self.col = 0

    def _clamp_col(self, buffer: Buffer) -> None:
        self._col = min(self._col_hint, len(buffer[self.row]))


class Window:
    def __init__(self, n_rows: int, n_cols: int, row: int = 0, col: int = 0) -> None:
        self.n_rows = n_rows
        self.n_cols = n_cols
        self.row = row
        self.col = col

    @property
    def bottom(self) -> int:
        return self.row + self.n_rows - 1

    def up(self, cursor: Cursor) -> None:
        if cursor.row == self.row - 1 and self.row > 0:
            self.row -= 1

    def down(self, buffer: Buffer, cursor: Cursor) -> None:
        if cursor.row == self.bottom + 1 and self.bottom < buffer.bottom:
            self.row += 1

    def translate(self, cursor: Cursor) -> Tuple[int, int]:
        return cursor.row - self.row, cursor.col - self.col

    def horizontal_scroll(
        self, cursor: Cursor, left_margin: int = 5, right_margin: int = 2
    ) -> None:
        n_pages = cursor.col // (self.n_cols - right_margin)
        self.col = max(n_pages * self.n_cols - right_margin - left_margin, 0)


def right(window: Window, buffer: Buffer, cursor: Cursor) -> None:
    cursor.right(buffer)
    window.down(buffer, cursor)
    window.horizontal_scroll(cursor)


def left(window: Window, buffer: Buffer, cursor: Cursor) -> None:
    cursor.left(buffer)
    window.up(cursor)
    window.horizontal_scroll(cursor)


class Editor:
    def __init__(
        self,
        stdscr: curses.window,
        value: str | None,
        validator: Callable[[int | str], int | str],
    ) -> None:
        self.validator = validator
        self.buffer = Buffer(value)
        self.lines, self.cols = stdscr.getmaxyx()
        self.stdscr = stdscr
        self.window = Window(self.lines - 1, self.cols - 2)
        self.cursor = Cursor()

    def set_value(self, lines: str) -> None:
        self.buffer.set_value(lines)

    def gather(self) -> str:
        return "\n".join(self.buffer.lines)

    def render(self) -> None:
        # TODO: Improve resize
        buffer = self.buffer
        window = self.window
        cursor = self.cursor
        self.stdscr.erase()
        if len(buffer.lines) > window.n_rows:
            y_pos = min(
                (window.row) / (len(buffer.lines) - window.n_rows) * window.n_rows,
                window.n_rows,
            )
            for row in range(window.n_rows + 1):
                if row == round(y_pos):
                    self.stdscr.addch(row, window.n_cols, "█")
                else:
                    self.stdscr.addch(row, window.n_cols, "│")

        for row, line in enumerate(buffer[window.row : window.row + window.n_rows]):
            if row == cursor.row - window.row and window.col > 0:
                line = "«" + line[window.col + 1 :]
            if len(line) > window.n_cols:
                self.stdscr.addstr(window.n_rows, 5, f"{len(line)}/{window.n_cols}")
                line = line[: window.n_cols - 1] + "»"
            self.stdscr.addstr(row, 0, line)
            self.stdscr.move(*window.translate(cursor))

    def main(self) -> None:
        buffer = self.buffer
        window = self.window
        self.stdscr.keypad(True)
        cursor = self.cursor
        curses.curs_set(True)
        while True:
            self.render()
            k = self.stdscr.get_wch()
            if not self.validator(k):
                break
            if k == "q":
                break
            elif k == curses.KEY_UP:
                cursor.up(buffer)
                window.up(cursor)
                window.horizontal_scroll(cursor)
            elif k == curses.KEY_DOWN:
                cursor.down(buffer)
                window.down(buffer, cursor)
                window.horizontal_scroll(cursor)
            elif k == curses.KEY_LEFT:
                cursor.left(buffer)
                window.up(cursor)
                window.horizontal_scroll(cursor)
            elif k == curses.KEY_RIGHT:
                cursor.right(buffer)
                window.down(buffer, cursor)
                window.horizontal_scroll(cursor)
            elif k == "\n":
                buffer.split(cursor)
                right(window, buffer, cursor)
            elif k == curses.KEY_BACKSPACE:
                if (cursor.row, cursor.col) > (0, 0):
                    left(window, buffer, cursor)
                    buffer.delete(cursor)
            elif k == curses.KEY_BTAB:
                buffer.insert(cursor, "  ")
                right(window, buffer, cursor)
                right(window, buffer, cursor)
            elif k == "\t":
                break
            elif isinstance(k, str):
                buffer.insert(cursor, k)
                for _ in k:
                    right(window, buffer, cursor)



from abc import ABC, abstractmethod
from curses import window
from typing import Any



class BaseView(ABC):
    def __init__(self, controler: Any) -> None:
        self.window: window = controler.window
        self.controller = controler

    @abstractmethod
    def render(self) -> None: ...

    @abstractmethod
    def main_loop(self) -> None: ...




import curses
from typing import Callable, Tuple


class Buffer:
    def __init__(self, value: str | None):
        self.line: str = value if value is not None else ""

    def set_value(self, line: str) -> None:
        self.line = line

    def insert(self, cursor: "Cursor", string: str) -> None:
        col = cursor.col
        self.line = self.line[:col] + string + self.line[col:]

    def delete(self, cursor: "Cursor") -> None:
        col = cursor.col
        if col < len(self.line):
            current = self.line
            if col < len(current):
                self.line = current[:col] + current[col + 1 :]


class Cursor:
    def __init__(self, row: int = 0, col: int = 0) -> None:
        self.row = row
        self._col = col

    @property
    def col(self) -> int:
        return self._col

    @col.setter
    def col(self, col: int) -> None:
        self._col = col

    def left(self) -> None:
        if self.col > 0:
            self.col -= 1

    def right(self, buffer: Buffer) -> None:
        if self.col < len(buffer.line):
            self.col += 1


class Window:
    def __init__(self, n_cols: int) -> None:
        self.n_cols = n_cols
        self.col = 0

    def translate(self, cursor: Cursor) -> Tuple[int, int]:
        return 0, cursor.col - self.col

    def horizontal_scroll(
        self, cursor: Cursor, left_margin: int = 5, right_margin: int = 0
    ) -> None:
        page_n_cols = self.n_cols - left_margin - right_margin
        n_pages = max((cursor.col - left_margin) // page_n_cols, 0)
        self.col = n_pages * page_n_cols


def right(window: Window, buffer: Buffer, cursor: Cursor) -> None:
    cursor.right(buffer)
    window.horizontal_scroll(cursor)


def left(window: Window, cursor: Cursor) -> None:
    cursor.left()
    window.horizontal_scroll(cursor)


class TextInput:
    def __init__(
        self,
        stdscr: curses.window,
        value: str | None,
        validator: Callable[[int | str], int | str],
    ) -> None:
        self.validator = validator
        self.buffer = Buffer(value)
        self.lines, self.cols = stdscr.getmaxyx()
        self.stdscr = stdscr
        self.window = Window(self.cols - 3)
        self.cursor = Cursor()

    def set_value(self, lines: str) -> None:
        self.buffer.set_value(lines)

    def gather(self) -> str:
        return "".join(self.buffer.line)

    def render(self) -> None:
        buffer = self.buffer
        window = self.window
        cursor = self.cursor
        self.stdscr.erase()

        line = buffer.line
        if window.col > 0:
            line = "«" + line[window.col + 1 :]
        if len(line) > window.n_cols:
            line = line[: window.n_cols - 1] + "»"
        self.stdscr.addstr(0, 0, line)
        self.stdscr.move(*window.translate(cursor))

    def main(self) -> None:
        buffer = self.buffer
        window = self.window
        self.stdscr.keypad(True)
        cursor = self.cursor
        cursor.col = min(len(self.buffer.line), window.n_cols)
        curses.curs_set(True)
        while True:
            self.render()
            k = self.stdscr.get_wch()
            if not self.validator(k):
                break
            if k == curses.KEY_LEFT:
                cursor.left()
                window.horizontal_scroll(cursor)
            elif k == curses.KEY_RIGHT:
                cursor.right(buffer)
                window.horizontal_scroll(cursor)
            elif k == "\n":
                pass
            elif k == curses.KEY_BACKSPACE:
                if (cursor.row, cursor.col) > (0, 0):
                    left(window, cursor)
                    buffer.delete(cursor)
            elif k == '\x0e' or k == '\x1b' or k == "\t":
                break
            elif isinstance(k, str):
                if k == "\t":
                    k = "  "
                buffer.insert(cursor, k)
                for _ in k:
                    right(window, buffer, cursor)



from __future__ import annotations

from curses import KEY_RESIZE
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict


from cursedtodo.config import Config
from cursedtodo.models.todo_repository import TodoRepository
from cursedtodo.utils.window_utils import add_borders, draw_line
from cursedtodo.views.base_view import BaseView
from cursedtodo.views.form.Button import Button
from cursedtodo.views.form.base_field import BaseField
from cursedtodo.views.form.categories_field import CategoriesField
from cursedtodo.views.form.checkbox_field import CheckboxField
from cursedtodo.views.form.datetime_field import DatetimeField
from cursedtodo.views.form.priority_field import PriorityField
from cursedtodo.views.form.select_field import SelectField
from cursedtodo.views.form.textarea_field import TextArea
from cursedtodo.views.form.textinput_field import TextInputField

if TYPE_CHECKING:
    from cursedtodo.controlers.edit_todo_controller import EditTodoController


class EditTodoView(BaseView):
    def __init__(self, controller: EditTodoController) -> None:
        super().__init__(controller)
        self.controller = controller
        self.height, self.length = self.window.getmaxyx()
        todo = self.controller.todo
        lists = TodoRepository.get_lists_names()
        default_list = Config.ui.default_calendar

        self.fields: Dict[str, BaseField] = {
            "list": SelectField(
                2, self.window, "List", "list", lists, self.validator, default_list
            ),
            "summary": TextInputField(
                3, self.window, "Summary", "summary", self.validator, ""
            ),
            "priority": PriorityField(
                4, self.window, "Priority", "priority", self.validator
            ),
            "due_checked": CheckboxField(
                5, self.window, "Due", "due_checked", self.validator
            ),
            "due": DatetimeField(6, self.window, "due", "", self.validator),
            "categories": CategoriesField(7, self.window, "categories", self.validator),
            "location": TextInputField(
                8, self.window, "Location", "location", self.validator
            ),
            "description": TextArea(
                10, self.window, "Description", "description", self.validator, ""
            ),
        }
        if todo:
            for key, field in self.fields.items():
                if key == "list":
                    field.value = todo.calendar.name
                elif key == "due_checked":
                    field.value = todo.due is not None
                elif key == "due":
                    field.value = todo.due or datetime.now()
                else:
                    field.value = getattr(todo, key)
        self.save_button = Button(
            self.window, self.height - 2, 1, "[ Save ]", self.save, self.validator
        )
        self.cancel_button = Button(
            self.window, self.height - 2, 10, "[ Cancel ]", self.cancel, self.validator
        )

    def render(self) -> None:
        self.height, self.length = self.window.getmaxyx()
        self.window.erase()
        self.window.border()
        add_borders(self.window)
        self.window.addstr(self.height - 1, 5, " tab: next field ")
        self.window.addstr(0, 5, " Edit todo ")
        draw_line(self.window, 9, self.length)
        for field in self.fields.values():
            field.render()
        self.save_button.render(self.height - 2, 1)
        self.cancel_button.render(self.height - 2, 10)
        self.window.refresh()

    def save(self) -> bool:
        values: Dict[str, Any] = {}
        for field in self.fields.values():
            values.update({field.id: field.value})
        self.controller.create_or_update_todo(values)
        return True

    def cancel(self) -> bool:
        return True

    def validator(self, ch: int | str) -> int | str:
        if ch == KEY_RESIZE:
            self.render()
        return ch

    def main_loop(self) -> None:
        index = 0
        fields = list(self.fields.values())
        self.window.refresh()
        while True:
            if index < len(self.fields):
                fields[index].focus()
                index += 1
            elif index == len(self.fields):
                if self.save_button.focus():
                    break
                index += 1
            elif index == len(self.fields) + 1:
                if self.cancel_button.focus():
                    break
                index += 1
            else:
                index = 0



from cursedtodo.controlers.base_controller import Controller
from curses import KEY_RESIZE

from cursedtodo.models.todo import Todo
from cursedtodo.views.view_todo_view import ViewTodoView

class ViewTodoController(Controller):



    def run(self, todo: Todo) -> None:
        self.todo = todo
        self.view = ViewTodoView(self)
        self.window.clear()
        self.view.render()
        self.view.main_loop()

    def handle_key(self, key: int) -> bool:
        if key == KEY_RESIZE:
            self.view.render()
        if key == ord("q"):
            return True
        return False



from collections.abc import Callable
from curses import A_BOLD, KEY_RESIZE, window
from datetime import datetime
from cursedtodo.utils.colors import WHITE
from cursedtodo.utils.textinput import TextInput
from cursedtodo.utils.time import TimeUtil
from cursedtodo.views.form.base_field import BaseField


class DatetimeField(BaseField):
    def __init__(
        self,
        y: int,
        window: window,
        id: str,
        name: str,
        validator: Callable[[int | str], int | str],
    ):
        super().__init__(y, window, name, id, validator)
        self.value: datetime | None = None
        self.textwindow = window.derwin(1, 20, y, 15)
        self.textwindow.bkgd(" ", WHITE)
        self.validator = validator
        self.editor = TextInput(self.textwindow, "", self._validator)

    def _validator(self, ch: str | int) -> str | int:
        if ch == KEY_RESIZE:
            self.value = TimeUtil.parse_to_datetime(self.editor.gather())
        self.validator(ch)
        return ch

    def render(self) -> None:
        self.textwindow.erase()
        if self.name != "":
            self.window.addstr(self.y, 1, f"{self.name}: ", A_BOLD)
        self.textwindow.move(0, 0)
        if self.value is not None:
            value = TimeUtil.datetime_format(self.value)
            self.editor.set_value(value)
        self.editor.render()
        self.textwindow.refresh()

    def focus(self) -> None:
        self.editor.main()
        value = self.editor.gather()
        if value is None or len(value) == 0:
            return
        self.value = TimeUtil.parse_to_datetime(self.editor.gather())



from argparse import ArgumentParser
from pathlib import Path
from dataclasses import dataclass


@dataclass
class _Arguments:
    config: Path
    verbose: bool


def _init_arg_parser() -> ArgumentParser:
    parser = ArgumentParser(description="CursedTodo")
    parser.add_argument("-c", "--config", type=Path, help="Path to the config file")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose output"
    )
    return parser


def _init_arguments() -> _Arguments:
    args = _init_arg_parser().parse_args()
    return _Arguments(config=args.config, verbose=args.verbose)


Arguments = _init_arguments()




from dataclasses import dataclass
import os
from typing import Optional

from cursedtodo.utils.colors import WHITE, get_color


@dataclass
class Calendar:
    id: int
    name: str
    path: str
    color: Optional[str] = "white"
    default: Optional[bool] = None
    color_attr: int = 0

    def __post_init__(self) -> None:
        self.color_attr = get_color(self.color)
        self.path = os.path.expanduser(self.path)



from typing import Any, List, TypeVar

T = TypeVar("T")


class Validator:
    @staticmethod
    def validate(value: Any, T: type[T]) -> T:
        if isinstance(value, T):
            return value
        else:
            raise Exception(f"{type(value).__name__} received, expected: {T.__name__}")

    @staticmethod
    def validate_optional(value: Any, T: type[T]) -> T | None:
        if value is None:
            return None
        if isinstance(value, T):
            return value
        else:
            raise Exception(f"{type(value).__name__} received, expected: {T.__name__}")

    @staticmethod
    def validate_list(value: Any, T: type[T]) -> List[T]:
        if not isinstance(value, list):
            raise TypeError(f"Expected a list, but got {type(value).__name__}.")

        if all(isinstance(item, T) for item in value):
            return value  

        raise ValueError(f"All items in the list must be of type {T.__name__}.")



from curses import window

from cursedtodo.controlers.edit_todo_controller import EditTodoController
from cursedtodo.controlers.main_controller import MainController
from cursedtodo.controlers.view_todo_controller import ViewTodoController
from cursedtodo.models.todo import Todo


class Router:
    def __init__(self, window: window) -> None:
        self.window = window

    def route_main(self) -> None:
        acontroller = MainController(self)
        acontroller.run()

    def route_view_todo(self, todo: Todo) -> None:
        ViewTodoController(self).run(todo)

    def route_create_todo(self) -> None:
        EditTodoController(self).run()

    def route_edit_todo(self, todo: Todo) -> None:
        EditTodoController(self).run(todo)
